#include "precomp.h"
#include "Enemy.h"
#include "World.h"
#include "Mario.h"

namespace Tmpl8 {

	Enemy::Enemy(int x, int y, int Type, bool Flying, bool Color) :	//constructor definition
		EnemyType(Type),
		EnemyAlive(true),
		EnemyFacingLeft(true),
		EnemyDead(false),
		EnemyDeathTimer(0),
		EnemyFlying(Flying),
		EnemyColorRed(Color),
		Kicked(false)
	{
		X = x;
		Y = y;
		SpeedY = 0;
	}

	void Enemy::ChangeEnemyDirection()
	{
		if (EnemyFacingLeft) EnemyFacingLeft = false;
		else EnemyFacingLeft = true;
	}

	void Enemy::EnemyMove(World* UseWorld, Mario* UseMario)
	{
		if ((EnemyAlive || EnemyType == 1) && SpeedY < 8) SpeedY += 0.3;

		if (EnemyType == 1) {
			if (EnemyAlive) {
				if (EnemyFacingLeft) X -= 1;
				else X += 1;
			}
			else if (Kicked) {
				if (EnemyFacingLeft) X -= 5;
				else X += 5;
			}
		}
		else {
			if (EnemyAlive) {
				if (EnemyFacingLeft) X -= 1;
				else X += 1;
			}
		}

		Y += SpeedY;

		CheckEnemyCollision(UseWorld, UseMario);
	}

	int Enemy::SetFrame(World* UseWorld)
	{
		bool EnemyFrame = UseWorld->GetEnemyFrame();
		int Frame = 0;

		//8 = Underground, 4 = Castle, 2 = UnderWater, 1 = Bonus
		unsigned char LevelFlags = UseWorld->GetLevelFlags();

		switch (EnemyType) {

		//Draw the goomba
		case 0:
			if (!(LevelFlags & (1 << 3 | 1 << 2 | 1 << 1))) Frame = 0;
			else if (LevelFlags & 1 << 2) Frame = 6;
			else if (LevelFlags & 1 << 3) Frame = 3;

			if (EnemyAlive) {
				if (EnemyFrame) Frame += 1;
			}
			else Frame += 2;

			if (!EnemyAlive) {
				++EnemyDeathTimer;
				if (EnemyDeathTimer == 70) EnemyDead = true;
			}
			break;

		//Draw the Koopa
		case 1:

			if (LevelFlags & 1 << 3 && !EnemyColorRed && !EnemyFlying) Frame = 4;
			else if (!EnemyColorRed && !EnemyFlying) Frame = 0;
			else if (!EnemyFlying) Frame = 8;
			else if (!EnemyColorRed) Frame = 12;
			else Frame = 14;

			if (EnemyAlive) {
				if (EnemyFrame) Frame += 1;
			}
			else {
				Frame += 2;
				if (!Kicked) {
					if (++EnemyDeathTimer == 500) {
						EnemyAlive = true;
						EnemyDeathTimer = 0;
					}
					else if (EnemyDeathTimer >= 280 && EnemyFrame) Frame += 1;
				}
				else EnemyDeathTimer = 0;
			}

			if (!EnemyFacingLeft) Frame += 16;
			break;
		}

		if (X < -800) EnemyDead = true;
		if (Y > SCRHEIGHT + 50) EnemyDead = true;

		if (EnemyDead) return -1;
		else return Frame;
	}

	void Enemy::CheckEnemyCollision(World* UseWorld, Mario* UseMario)
	{
		int WorldXOffset = UseWorld->GetWorldXOffset();
		int WorldX = UseWorld->GetWorldX();

		int StartingRow = floor(WorldX / 80) - 10;
		int EndingRow = StartingRow + 16 + 100;
		int BlockX;
		int BlockY;
		int Block;
		int MarioX = UseMario->GetX();
		int MarioY = UseMario->GetY();
		int MarioSpeedY = UseMario->GetSpeedY();
		int MarioHeight;
		int MarioSize = UseMario->GetMarioSize();
		bool MarioInvincibility = UseMario->GetInvincibility();
		bool MarioStar = UseMario->GetStar();

		if (MarioSize == 0) {
			MarioY = MarioY - 80;
			MarioHeight = 80;
		}
		else {
			MarioY = MarioY - 160;
			MarioHeight = 160;
		}

		//Check collision with Mario
		if (MarioStar) {
			if (MarioX < X + 80 && MarioX + 80 > X && MarioY < Y + 80 && MarioY + MarioHeight > Y) {
				EnemyDead = true;
				switch (EnemyType) {
				case 0:
					UseWorld->AddScore(100);
					break;
				case 1:
					UseWorld->AddScore(200);
					break;
				}
			}
		}
		else {
			if (EnemyAlive) {
				if (MarioX < X + 80 && MarioX + 80 > X && MarioY < Y + 80 && MarioY + MarioHeight > Y) {

					//top hitbox
					if (MarioX < (X + 10) + 60 && MarioX + 80 >(X + 10) && MarioY < Y && MarioY + MarioHeight > Y) {
						if (MarioSpeedY > 0) {
							EnemyAlive = false;

							UseMario->SetMarioSpeedY(-5);
							UseMario->SetY(Y);
							UseMario->SetMarioJump(true);
							UseMario->SetMarioJumpEnd(true);
							UseWorld->AddScore(100);
						}
					}

					//left hitbox
					else if (MarioX < X && MarioX + 80 > X && MarioY < (Y + 10) + 60 && MarioY + MarioHeight >(Y + 10)) {
						if (!MarioInvincibility) {
							if (MarioSize == 1 || MarioSize == 2) UseMario->RemoveMarioSize();
							else UseMario->SetMarioAlive(false);
						}
					}

					//right hitbox
					else if (MarioX < (X + 80) && MarioX + 80 >(X + 80) && MarioY < (Y + 10) + 60 && MarioY + MarioHeight >(Y + 10)) {
						if (!MarioInvincibility) {
							if (MarioSize == 1 || MarioSize == 2) UseMario->RemoveMarioSize();
							else UseMario->SetMarioAlive(false);
						}
					}
				}
			}
			else if (!EnemyAlive && EnemyType == 1) {
				if (MarioX < X + 80 && MarioX + 80 > X && MarioY < Y + 80 && MarioY + MarioHeight > Y) {
					//top hitbox
					if (MarioX < (X + 10) + 60 && MarioX + 80 >(X + 10) && MarioY < Y && MarioY + MarioHeight > Y) {
						if (MarioSpeedY > 0) {
							if (!Kicked) {
								if (MarioX < X) {
									Kicked = true;
									EnemyFacingLeft = false;
								}
								else if (MarioX + 80 > X + 80) {
									Kicked = true;
									EnemyFacingLeft = true;
								}
							}
							else Kicked = false;

							UseMario->SetMarioSpeedY(-5);
							UseMario->SetY(Y);
							UseMario->SetMarioJump(true);
							UseMario->SetMarioJumpEnd(true);
						}
					}

					//left hitbox
					else if (MarioX < X && MarioX + 80 > X && MarioY < (Y + 10) + 60 && MarioY + MarioHeight >(Y + 10)) {
						if (!Kicked) {
							Kicked = true;
							EnemyFacingLeft = false;
						}
						else if (!MarioInvincibility) {
							if (MarioSize == 1 || MarioSize == 2) UseMario->RemoveMarioSize();
							else UseMario->SetMarioAlive(false);
						}
					}

					//right hitbox
					else if (MarioX < (X + 80) && MarioX + 80 >(X + 80) && MarioY < (Y + 10) + 60 && MarioY + MarioHeight >(Y + 10)) {
						if (!Kicked) {
							Kicked = true;
							EnemyFacingLeft = true;
						}
						else if (!MarioInvincibility) {
							if (MarioSize == 1 || MarioSize == 2) UseMario->RemoveMarioSize();
							else UseMario->SetMarioAlive(false);
						}
					}
				}
			}
		}

		//Check collision with blocks
		for (int y = 0; y < 13; y++) {
			for (int x = 0; x < 213; x++) {
				if (x >= StartingRow && x <= EndingRow) {

					Block = UseWorld->GetBlock(x, y);

					if (Block > 0 && Block != 17) {
						BlockX = (x * 80) - (WorldX - (WorldX % 80) - -WorldXOffset);
						BlockY = y * 80;

						//Check if the enemy even hits a block
						if (X < BlockX + 80 && X + 80 > BlockX && Y < BlockY + 80 && Y + 80 > 80) {

							//top hitbox
							if (X < (BlockX + 10) + 60 && X + 80 >(BlockX + 10) && Y < BlockY && Y + 80 > BlockY) {
								Y = BlockY - 80;
								SpeedY = 0;
							}

							//left hitbox
							else if (X < BlockX && X + 80 > BlockX && Y < (BlockY + 10) + 60 && Y + 80 >(BlockY + 10)) {
								X = BlockX - 80;
								EnemyFacingLeft = true;
							}

							//right hitbox
							else if (X < (BlockX + 80) && X + 80 >(BlockX + 80) && Y < (BlockY + 10) + 60 && Y + 80 >(BlockY + 10)) {
								X = BlockX + 80;
								EnemyFacingLeft = false;
							}
						}
					}
				}
			}
		}
	}
}