#include "precomp.h"
#include "Mario.h"
#include "game.h"
#include "World.h"

namespace Tmpl8 {

	Mario::Mario() :
		MarioSize(0),
		Invincibility(0),
		MarioFrame(0),
		MarioWalkFrame(1),
		MarioWalkTimer(1),
		MarioJumpTimer(30),
		MarioDeadTimer(0),
		FlagTimer(1),
		EnterPipeTimer(0),
		EnterVPipe(0),
		EnterHPipe(0),
		StarTimer(0),
		StarFrame(0),

		MarioAlive(true),
		MarioFacingLeft(false),
		MarioBoostLeft(false),
		MarioBoostRight(false),
		MarioWalk(false),
		MarioJump(false),
		MarioJumpStart(false),
		MarioJumpEnd(false),
		FlagFrame(false),
		ExitPipe(false),
		HPipe(false),
		VPipe(false),
		Star(false)
	{
		for (int i = 0; i < 5; i++) {
			Movement[i] = false;
		}

		X = 160;
		Y = 960;
		SpeedX = 0;
		SpeedY = 0;
	}

	void Mario::Move(World* UseWorld, int FinishedTimer)
	{
		if (Invincibility > 0) --Invincibility;
		if (Star) {
			if (--StarTimer == 0) {
				Star = false;
				StarFrame = 0;
			}
			if (StarTimer % 10 == 0) {
				if (++StarFrame == 4) StarFrame = 0;
			}
		}

		//automatically move mario into the pipe in 1-2
		if (UseWorld->GetLevel() == 3 && UseWorld->GetWorldX() < 300) {
			for (int i = 0; i < 5; i++) {
				Movement[i] = 0;
			}
			Movement[2] = true;
		}

		if (FinishedTimer > 0) FinishAnimation(UseWorld, FinishedTimer);
		else if (ExitPipe) {
			if (MarioSize == 0) Y -= 1;
			else Y -= 2;

			if (Y <= 800) ExitPipe = false;

			MarioFrame = 0;
			if (MarioSize == 1) MarioFrame += 14;
			else if (MarioSize == 2) MarioFrame += 28;
		}
		else if (VPipe) {
			if (++EnterPipeTimer < 160) {
				if (MarioSize == 0) Y += 1;
				else Y += 2;
			}
			else {
				EnterPipeTimer = 0;
				VPipe = false;
				UseWorld->EnterPipe();
			}

			MarioFrame = 0;
			if (MarioFacingLeft) MarioFrame += 42;

			if (MarioSize == 1) MarioFrame += 14;
			else if (MarioSize == 2) MarioFrame += 28;
		}
		else if (HPipe) {
			if (++EnterPipeTimer < 160) ++X;
			else {
				EnterPipeTimer = 0;
				HPipe = false;
				UseWorld->EnterPipe();
			}

			MarioFrame = 0;
			if (MarioFacingLeft) MarioFrame += 42;

			if (MarioSize == 1) MarioFrame += 14;
			else if (MarioSize == 2) MarioFrame += 28;

			if (--MarioWalkTimer == 0) {
				MarioWalkTimer = 15;
				if (--MarioWalkFrame == 0) MarioWalkFrame = 3;
			}
			MarioFrame += MarioWalkFrame;
		}
		else {
			EnterVPipe = 0;
			EnterHPipe = 0;

			if (!MarioAlive) {
				if (++MarioDeadTimer == 1) {
					SpeedX = 0;
					SpeedY = -8;
					MarioSize = 0;
					MarioFrame = 6;
				}
				else if (MarioDeadTimer > 50 && MarioDeadTimer < 200) {
					if (SpeedY < 8) {
						SpeedY += 0.2;
					}
				}
				else if (MarioDeadTimer == 200) {
					UseWorld->RemoveLife();
					UseWorld->Reset();
				}
				else if (MarioDeadTimer == 201) {
					MarioReset(UseWorld->GetLevel());
				}

				Y += SpeedY;
			}
			else {
				//x-axis
				if (Movement[1] && Movement[2]) {

					//Stop Mario from moving
					if (SpeedX > 0.1) SpeedX -= 0.1;
					if (SpeedX < -0.1) SpeedX += 0.1;
					if (SpeedX < 0.1 && SpeedX > -0.1) {
						SpeedX = 0;
						MarioWalk = false;
					}
				}
				else {
					//Move Mario left or right
					if (Movement[4]) {
						if (Movement[2] && SpeedX < 6) SpeedX += 0.03;
						if (Movement[1] && SpeedX > -6) SpeedX -= 0.03;
					}
					else {
						if (Movement[2] && SpeedX < 3) SpeedX += 0.03;
						if (Movement[1] && SpeedX > -3) SpeedX -= 0.03;

						if (SpeedX > 3) SpeedX -= 0.05;
						if (SpeedX < -3) SpeedX += 0.05;
					}

					//Slow down Mario if user is pressing no button
					if (!Movement[2] && SpeedX > 0.1) SpeedX -= 0.07;
					if (!Movement[1] && SpeedX < -0.1) SpeedX += 0.07;
					if (!Movement[1] && !Movement[2] && SpeedX < 0.1 && SpeedX > -0.1) {
						SpeedX = 0;
						MarioWalk = false;
					}

					//Give Mario a boost when turning around
					if (SpeedX > 1 && Movement[1]) MarioBoostLeft = true;
					if (SpeedX < -1 && Movement[2]) MarioBoostRight = true;
					if (MarioBoostLeft && Movement[2]) MarioBoostLeft = false;
					if (MarioBoostRight && Movement[1]) MarioBoostRight = false;
					if (!Movement[1] && !Movement[2]) {
						MarioBoostLeft = false;
						MarioBoostRight = false;
					}
					if (MarioBoostLeft) {
						if (SpeedX < -1) {
							MarioBoostLeft = false;
						}
						else if (SpeedX < 0) {
							SpeedX -= 0.1;
						}
					}
					if (MarioBoostRight) {
						if (SpeedX > 1) {
							MarioBoostRight = false;
						}
						else if (SpeedX > 0) {
							SpeedX += 0.1;
						}
					}

					//Check collision with the left wall
					if (X < 0) {
						X = 0;
						SpeedX = 0;
					}

					//Setting Mario's frame
					if (Movement[2]) MarioFacingLeft = false;
					if (Movement[1]) MarioFacingLeft = true;
					if (Movement[1] || Movement[2]) MarioWalk = true;
				}

				//y-axis
				if (Movement[0] && !MarioJumpEnd) {
					SpeedY = -8.5;
					MarioJump = true;
				}

				if (SpeedY != 0) MarioJumpStart = true;

				if (!Movement[0] && MarioJumpStart) MarioJumpEnd = true;
				if (MarioJumpStart) {
					if (--MarioJumpTimer == 0) MarioJumpEnd = true;
				}
				if (!Movement[0] && !MarioJump) {
					MarioJumpStart = false;
					MarioJumpEnd = false;
					MarioJumpTimer = 30;
				}

				if (SpeedY > 1) MarioJumpEnd = true;

				if (SpeedY < 8) SpeedY += 0.3;


				//Move Mario
				X += SpeedX;
				Y += SpeedY;

				CheckMarioCollision(UseWorld);

				//Vertical Pipe
				if (EnterVPipe == 2 && Movement[3]) {
					VPipe = true;

					MarioFrame = 0;
					if (MarioFacingLeft) MarioFrame += 42;

					if (MarioSize == 1) MarioFrame += 14;
					else if (MarioSize == 2) MarioFrame += 28;
				}

				//Horizontal Pipe
				else if (EnterHPipe == 2 && Movement[2]) {
					HPipe = true;

					MarioFrame = 0;
					if (MarioFacingLeft) MarioFrame += 42;

					if (MarioSize == 1) MarioFrame += 14;
					else if (MarioSize == 2) MarioFrame += 28;

					if (--MarioWalkTimer == 0) {
						if (SpeedX <= 2 && SpeedX >= -2) MarioWalkTimer = 15;
						else if ((SpeedX > 2 && SpeedX <= 4) || (SpeedX < -2 && SpeedX >= -4)) MarioWalkTimer = 10;
						else MarioWalkTimer = 7;

						if (--MarioWalkFrame == 0) MarioWalkFrame = 3;
					}
					MarioFrame += MarioWalkFrame;
				}

				//Star frames
				else if (Star && (StarFrame == 1 || StarFrame == 3)) {
					MarioFrame = 100;

					if (StarFrame == 3) MarioFrame += 56;

					if (MarioFacingLeft) MarioFrame += 28;
					if (MarioSize != 0) MarioFrame += 14;

					if (MarioJump) MarioFrame += 5;
					else if (MarioBoostLeft || MarioBoostRight) MarioFrame += 4;
					else if (MarioWalk) {
						if (--MarioWalkTimer == 0) {
							if (SpeedX <= 2 && SpeedX >= -2) MarioWalkTimer = 15;
							else if ((SpeedX > 2 && SpeedX <= 4) || (SpeedX < -2 && SpeedX >= -4)) MarioWalkTimer = 10;
							else MarioWalkTimer = 7;

							if (--MarioWalkFrame == 0) MarioWalkFrame = 3;
						}
						MarioFrame += MarioWalkFrame;
					}
					else if (Movement[3] && SpeedX == 0 && MarioSize > 0) MarioFrame += 6;
				}

				else {
					//Set frame of Mario
					MarioFrame = 0;
					if (MarioFacingLeft) MarioFrame += 42;

					if (MarioSize == 1) MarioFrame += 14;
					else if (MarioSize == 2) MarioFrame += 28;

					if (MarioJump) MarioFrame += 5;
					else if (MarioBoostLeft || MarioBoostRight) MarioFrame += 4;
					else if (MarioWalk) {
						if (--MarioWalkTimer == 0) {
							if (SpeedX <= 2 && SpeedX >= -2) MarioWalkTimer = 15;
							else if ((SpeedX > 2 && SpeedX <= 4) || (SpeedX < -2 && SpeedX >= -4)) MarioWalkTimer = 10;
							else MarioWalkTimer = 7;

							if (--MarioWalkFrame == 0) MarioWalkFrame = 3;
						}
						MarioFrame += MarioWalkFrame;
					}
					else if (Movement[3] && SpeedX == 0 && MarioSize > 0) MarioFrame += 6;
				}
			}
		}
	}

	void Mario::FinishAnimation(World* UseWorld, int FinishedTimer)
	{
		MarioFrame = 0;
		if (MarioSize == 1) MarioFrame += 14;
		else if (MarioSize == 2) MarioFrame += 28;

		if (FinishedTimer == 1) {
			UseWorld->AddScore(2000);
		}
		
		if (FinishedTimer > 0 && FinishedTimer <= 200) {
			MarioWalkTimer = 1;
			MarioWalkFrame = 1;
			SpeedX = 0;
			SpeedY = 0;

			if (Y < 880) {
				Y += 3;
				if (--FlagTimer == 0) {
					FlagTimer = 20;
					FlagFrame = !FlagFrame;
				}
			}
			else Y = 880;

			if (FlagFrame) MarioFrame += 8;
			else MarioFrame += 7;
		}
		else if (FinishedTimer > 200 && FinishedTimer <= 280) {
			X = 660;
			MarioFrame += 42;
			if (FlagFrame) MarioFrame += 8;
			else MarioFrame += 7;
		}
		else if (FinishedTimer > 280 && FinishedTimer <= 420) {
			if (SpeedX < 3) SpeedX += 0.3;
			if (SpeedY < 8) SpeedY += 0.3;

			X += SpeedX;
			Y += SpeedY;

			CheckMarioCollision(UseWorld);

			if (--MarioWalkTimer == 0) {
				if (SpeedX <= 2) MarioWalkTimer = 15;
				else if (SpeedX > 2 && SpeedX <= 4) MarioWalkTimer = 10;
				else MarioWalkTimer = 7;

				if (--MarioWalkFrame == 0) MarioWalkFrame = 3;
			}
			MarioFrame += MarioWalkFrame;
		}
		else UseWorld->SetFinished(true);
	}

	void Mario::CheckMarioCollision(World* UseWorld)
	{
		if (Y > SCRHEIGHT + 80) {
			MarioAlive = false;
			MarioSize = 0;
		}
		else {
			int WorldXOffset = UseWorld->GetWorldXOffset();
			int WorldX = UseWorld->GetWorldX();

			int StartingRow = floor(WorldX / 80);
			int EndingRow = StartingRow + 16;
			int BlockX;
			int BlockY;
			int marioX = X;
			int marioY;
			int marioHeight;
			int Block;

			if (MarioSize == 0) {
				marioY = Y - 80;
				marioHeight = 80;
			}
			else {
				marioY = Y - 160;
				marioHeight = 160;
			}

			//Check collision with blocks
			for (int y = 0; y < 13; y++) {
				for (int x = 0; x < 213; x++) {
					if (x >= StartingRow && x <= EndingRow) {

						Block = UseWorld->GetBlock(x, y);

						if (Block != 0) {
							BlockX = (x * 80) - (WorldX - (WorldX % 80) - -WorldXOffset);
							BlockY = y * 80;

							//Check if you even hit the block
							if (marioX < BlockX + 80 && marioX + 80 > BlockX && marioY < BlockY + 80 && marioY + marioHeight > BlockY) {


								switch (Block) {

								//Brick and QuestionBlock
								case 2: case 3: case 24: case 41:
									//top hitbox
									if (marioX < (BlockX + 10) + 60 && marioX + 80 >(BlockX + 10) && marioY < BlockY && marioY + marioHeight > BlockY) {
										Y = BlockY;
										SpeedY = 0;
										MarioJump = false;
									}

									//left hitbox
									else if (marioX < BlockX && marioX + 80 > BlockX && marioY < (BlockY + 10) + 60 && marioY + marioHeight >(BlockY + 10)) {
										X = BlockX - 80;
										SpeedX = 0;
									}

									//right hitbox
									else if (marioX < (BlockX + 80) && marioX + 80 >(BlockX + 80) && marioY < (BlockY + 10) + 60 && marioY + marioHeight >(BlockY + 10)) {
										X = BlockX + 80;
										SpeedX = 0;
									}

									//bottom hitbox
									else if (marioX < (BlockX + 10) + 60 && marioX + 80 >(BlockX + 10) && marioY < (BlockY + 80) && marioY + marioHeight >(BlockY + 80)) {

										//Check which block Mario exactly hit
										if ((marioX + 38) < BlockX + 60 && (marioX + 38) + 2 > BlockX && marioY < (BlockY + 80) && marioY + marioHeight >(BlockY + 80)) {
											if (Block == 3) {
												UseWorld->SetBlock(x, y, Block + 100);
												UseWorld->AddCoin(BlockX + 20, BlockY + 5);
											}
											else if (Block == 2) {
												if (MarioSize > 0) {
													UseWorld->SetBlock(x, y, Block - 100);
												}
											}
											else if (Block == 24) {
												UseWorld->SetBlock(x, y, Block + 100);
												UseWorld->AddPowerUp(BlockX, BlockY, 1);
											}
											else if (Block == 41) {
												UseWorld->SetBlock(x, y, Block + 100);
												UseWorld->AddPowerUp(BlockX, BlockY, 3);
											}
										}

										Y = BlockY + 80 + marioHeight;
										SpeedY = 0;
										MarioJumpEnd = true;
									}
									break;

								//Coin
								case 17:
									if (marioX < (BlockX + 15) + 50 && marioX + 80 >(BlockX + 15) && marioY < (BlockY + 10) + 70 && marioY + marioHeight >(BlockY + 10)) {
										UseWorld->SetBlock(x, y, Block - 100);
										UseWorld->AddCoin(NULL, NULL);
									}
									break;

								//Hidden block
								case 18: case 37:
									if (marioX < (BlockX + 10) + 60 && marioX + 80 >(BlockX + 10) && marioY < (BlockY + 80) && marioY + marioHeight >(BlockY + 80)) {
										if (SpeedY < 0) {
											Y = BlockY + 80 + marioHeight;
											SpeedY = 0;
											MarioJumpEnd = true;
											if (Block == 18) {
												UseWorld->SetBlock(x, y, 100);	//Set to 100 instead of block + 100
												UseWorld->AddPowerUp(BlockX, BlockY, 2);
											}
											else {
												UseWorld->SetBlock(x, y, Block + 100);
												UseWorld->AddCoin(BlockX + 20, BlockY + 5);
											}
										}
									}
									break;

								//Platforms
								case 25: case 26: case 27: case 28:
									if (SpeedY > 0) {
										int LeftOffset = 0;
										int RightOffset = 0;

										if (Block == 26) LeftOffset = 60;
										else if (Block == 28) RightOffset = 60;

										//top hitbox
										if (marioX < BlockX + 80 - RightOffset && marioX + 80 > BlockX + LeftOffset && (marioY + marioHeight) < (BlockY + 20) + 10 && marioY + marioHeight > (BlockY + 20)) {
											Y = BlockY + 20;
											SpeedY = 0;
											MarioJump = false;
										}
									}
									break;

								//Castle Axe
								case 39:
									if (marioX < (BlockX - 10) + 80 && marioX + 80 >(BlockX + 10) && marioY < (BlockY - 10) + 80 && marioY + marioHeight >(BlockY + 10)) {
										UseWorld->NextLevel();
									}
									break;

								default:
									if (Block > 0) {
										//top hitbox
										if (marioX < (BlockX + 10) + 60 && marioX + 80 >(BlockX + 10) && marioY < BlockY && marioY + marioHeight > BlockY) {
											Y = BlockY;
											SpeedY = 0;
											MarioJump = false;
										}

										//left hitbox
										else if (marioX < BlockX && marioX + 80 > BlockX && marioY < (BlockY + 10) + 60 && marioY + marioHeight >(BlockY + 10)) {
											X = BlockX - 80;
											SpeedX = 0;
										}

										//right hitbox
										else if (marioX < (BlockX + 80) && marioX + 80 >(BlockX + 80) && marioY < (BlockY + 10) + 60 && marioY + marioHeight >(BlockY + 10)) {
											X = BlockX + 80;
											SpeedX = 0;
										}

										//bottom hitbox
										else if (marioX < (BlockX + 10) + 60 && marioX + 80 >(BlockX + 10) && marioY < (BlockY + 80) && marioY + marioHeight >(BlockY + 80)) {
											Y = BlockY + 80 + marioHeight;
											SpeedY = 0;
											MarioJumpEnd = true;
										}


										//Enter VPipe
										if (Block == 20 || Block == 21) ++EnterVPipe;

										//Enter HPipe
										if (Block == 22 || Block == 23) ++EnterHPipe;
									}
								}								
							}
						}
					}
				}
			}
		}
	}

	void Mario::MarioReset(int Level)
	{
		X = 160;
		if (Level == 7) Y = 480;
		else Y = 960;

		SpeedX = 0;
		SpeedY = 0;
		MarioFrame = 0;
		MarioWalkFrame = 1;
		MarioWalkTimer = 1;
		MarioJumpTimer = 30;
		MarioDeadTimer = 0;
		MarioAlive = true;
		MarioFacingLeft = false;
		MarioBoostLeft = false;
		MarioBoostRight = false;
		MarioWalk = false;
		MarioJump = false;
		MarioJumpStart = false;
		MarioJumpEnd = false;
		Star = false;
	}

	void Mario::SetMovement(bool Jump, bool Left, bool Right, bool Duck, bool Sprint)
	{
		Movement[0] = Jump;
		Movement[1] = Left;
		Movement[2] = Right;
		Movement[3] = Duck;
		Movement[4] = Sprint;
	}

}